<?php

namespace Drupal\commerce_forteecom\Plugin\Commerce\PaymentGateway;

use Drupal;
use Drupal\commerce_forteecom\ForteEcomHelper;
use Drupal\commerce_order\Entity\Order;
use Drupal\commerce_order\Entity\OrderInterface;
use Drupal\commerce_order\OrderStorage;
use Drupal\commerce_payment\CreditCard;
use Drupal\commerce_payment\Entity\Payment;
use Drupal\commerce_payment\Entity\PaymentInterface;
use Drupal\commerce_payment\Entity\PaymentMethod;
use Drupal\commerce_payment\Entity\PaymentMethodInterface;
use Drupal\commerce_payment\Exception\PaymentGatewayException;
use Drupal\commerce_payment\PaymentMethodStorageInterface;
use Drupal\commerce_payment\PaymentStorageInterface;
use Drupal\commerce_payment\Plugin\Commerce\PaymentGateway\OffsitePaymentGatewayBase;
use Drupal\commerce_payment\Plugin\Commerce\PaymentGateway\SupportsAuthorizationsInterface;
use Drupal\commerce_payment\Plugin\Commerce\PaymentGateway\SupportsRefundsInterface;
use Drupal\commerce_price\Price;
use Drupal\Core\Form\FormStateInterface;
use InvalidArgumentException;
use Symfony\Component\HttpFoundation\Request;

/**
 * Provides the ForteEcom offsite Checkout payment gateway.
 *
 * If implements SupportsStoredPaymentMethodsInterface, gateway can store credit cards in account, but it forces
 * gateway
 * to enter credit card data on order_information page. Currently, only implemented for onsite gateways. @link
 * https://docs.drupalcommerce.org/commerce2/developer-guide/payments
 *
 * @CommercePaymentGateway(
 *   id = "forteecom_checkout",
 *   label = @Translation("Forteecom"),
 *   display_label = @Translation("ForteEcom"),
 *   forms = {
 *     "offsite-payment" = "Drupal\commerce_forteecom\PluginForm\ForteEcomCheckoutForm",
 *     "add-payment-method" = "Drupal\commerce_forteecom\PluginForm\PaymentMethodAddForm",
 *   },
 *   payment_method_types = {"credit_card"},
 *   payment_type = "payment_forteecom",
 *   credit_card_types = {
 *     "mastercard", "visa",
 *   },
 *   requires_billing_information = FALSE,
 * )
 */
class ForteEcomCheckout extends OffsitePaymentGatewayBase implements SupportsAuthorizationsInterface, SupportsRefundsInterface
{

  /**
   * @inheritDoc
   */
  public function defaultConfiguration()
  {
    return [
        'merchant_id' => '',
        'pre_auth'    => false,
      ] + parent::defaultConfiguration();
  }

  /**
   * @inheritDoc
   */
  public function buildConfigurationForm(array $form, FormStateInterface $form_state)
  {
    $form = parent::buildConfigurationForm($form, $form_state);

    $form['merchant_id'] = [
      '#type'          => 'textfield',
      '#title'         => $this->t('Merchant ID'),
      '#description'   => $this->t('Paste here your merchant ID from ForteEcom'),
      '#default_value' => $this->configuration['merchant_id'],
      '#required'      => TRUE,
    ];

    $form['pre_auth'] = [
      '#type'          => 'checkbox',
      '#title'         => $this->t('On-hold state'),
      '#description'   => $this->t('If enabled, payments firstly go to on-hold state. You will need to manually CAPTURE payment.'),
      '#default_value' => $this->configuration['pre_auth'],
    ];

    return $form;
  }

  /**
   * @inheritDoc
   */
  public function submitConfigurationForm(array &$form, FormStateInterface $form_state)
  {
    parent::submitConfigurationForm($form, $form_state);
    $values                             = $form_state->getValue($form['#parents']);
    $this->configuration['merchant_id'] = $values['merchant_id'];
    $this->configuration['pre_auth']    = $values['pre_auth'];
  }

  /**
   * @inheritDoc
   */
  public function onNotify(Request $request)
  {
    define('FORTEECOM_DEMO_ENABLED', ($this->getMode() == 'test') ? 1 : 0);

    $request_body = json_decode(file_get_contents( "php://input" ), true);

    $type = $request_body['type'];
    $data = $request_body['data'];

    switch ($type) {
      case "CardSaved":
        /** @var PaymentMethodStorageInterface $payment_method_storage */
        $payment_method_storage = $this->entityTypeManager->getStorage('commerce_payment_method');
        /** @var PaymentMethod $payment_method */
        $existing_payment_method = $payment_method_storage->loadByProperties([
          'type'            => 'credit_card',
          'payment_gateway' => $data['payment_gateway'],
          'uid'             => Drupal::currentUser()->id(),
          'remote_id'       => $data['card_token']
        ]);

        if (count($existing_payment_method) > 0) {
          return;
        }

        $payment_method = $payment_method_storage->create([
          'type'            => 'credit_card',
          'payment_gateway' => $data['payment_gateway'],
          'uid'             => Drupal::currentUser()->id(),
        ]);

        $payment_details = [
          'forteecom_token' => $data['card_token'],
          'type'            => $data['card_type'],
          'number'          => $data['card_last4'],
          'expiration'      => [
            'month' => $data['card_expiry_month'],
            'year'  => $data['card_expiry_year'],
          ]
        ];

        $this->createPaymentMethod($payment_method, $payment_details);

        break;
      case "PaymentComplete":

        $_forteecom_order_id   = $data['orderId'];
        $_forteecom_session_id = $data['sessionId'];
        $_forteecom_remote_id  = $data['remoteId'];

        /**
         * @var OrderStorage            $order_storage
         * @var PaymentStorageInterface $payment_storage
         * @var Order                   $order
         */
        $order_storage   = $this->entityTypeManager->getStorage('commerce_order');
        $payment_storage = $this->entityTypeManager->getStorage('commerce_payment');
        $order           = $order_storage->load($_forteecom_remote_id);
        $payment         = $payment_storage->loadByRemoteId($_forteecom_order_id);

        $_forteecom_merchant_id = $order->getData('commerce_forteecom')['_forteecom_merchant_id'];

        $status_result = ForteEcomHelper::get_payment_status($_forteecom_merchant_id, $_forteecom_order_id, $_forteecom_session_id);

        if (empty($status_result['error_message'])) {
          ForteEcomHelper::update_payment_status($order, $payment, $status_result['_forteecom_order_status']);
        }

        break;
      case "PaymentRetried":

        $_forteecom_order_id   = $data['orderId'];
        $_forteecom_session_id = $data['sessionId'];
        $_forteecom_remote_id  = $data['remoteId'];

        /**
         * @var OrderStorage            $order_storage
         * @var PaymentStorageInterface $payment_storage
         * @var Order                   $order
         */
        $order_storage   = $this->entityTypeManager->getStorage('commerce_order');
        $payment_storage = $this->entityTypeManager->getStorage('commerce_payment');
        $order           = $order_storage->load($_forteecom_remote_id);
        $payment         = $payment_storage->loadByRemoteId($_forteecom_order_id);


        $_forteecom_merchant_id = $order->getData('commerce_forteecom')['_forteecom_merchant_id'];
        $payment_gateway_plugin = $payment->getPaymentGateway()->getPlugin();
        $configuration          = $payment_gateway_plugin->getConfiguration();

        $status_result = ForteEcomHelper::get_payment_status($_forteecom_merchant_id, $_forteecom_order_id, $_forteecom_session_id);
        if (in_array($status_result['_forteecom_order_status'], [ForteEcomHelper::STATUS_DECLINE, ForteEcomHelper::STATUS_EXPIRE, ForteEcomHelper::STATUS_CANCEL, ForteEcomHelper::STATUS_ERROR])) {

          $init_result = ForteEcomHelper::init_payment(
            $configuration['merchant_id'], 
            $configuration['pre_auth'], 
            $payment->getAmount()->getNumber(), 
            $payment->getOrderId(), 
            Drupal::config('system.site')->get('name') . ' - Order: ' . $order->id(), 
            $payment->getAmount()->getCurrencyCode()
          );

          ForteEcomHelper::update_order_forteecom_data(
            $order,
            $init_result['_forteecom_order_id'],
            ForteEcomHelper::STATUS_INIT,
            $init_result['_forteecom_session_id'],
            $configuration['merchant_id']
          );

          $order->save();
          $payment->setRemoteId($init_result['_forteecom_order_id']);
          $payment->setState('inited');
          $payment->setRemoteState(ForteEcomHelper::STATUS_INIT);
          $payment->save();

          $response = [
            "sessionId" => $init_result['_forteecom_session_id'],
            "orderId"   => $init_result['_forteecom_order_id']
          ];
        } else {
          $response = [
            "sessionId" => $_forteecom_session_id,
            "orderId"   => $_forteecom_order_id
          ];
        }
        echo json_encode($response);
        break;
    }
  }

  /**
   * @inheritdoc
   * @throws Drupal\Core\Entity\EntityStorageException
   */
  public function createPaymentMethod(PaymentMethodInterface $payment_method, array $payment_details)
  {

    $required_keys = ['forteecom_token', 'type', 'number', 'expiration'];
    foreach ($required_keys as $required_key) {
      if (empty($payment_details[$required_key])) {
        throw new InvalidArgumentException(sprintf('$payment_details must contain the %s key.', $required_key));
      }
    }

    $payment_method->card_type      = $payment_details['type'];
    $payment_method->card_number    = $payment_details['number'];
    $payment_method->card_exp_month = $payment_details['expiration']['month'];
    $payment_method->card_exp_year  = $payment_details['expiration']['year'];
    $expires                        = CreditCard::calculateExpirationTimestamp($payment_details['expiration']['month'], $payment_details['expiration']['year']);

    $payment_method->setRemoteId($payment_details['forteecom_token']);
    $payment_method->setExpiresTime($expires);
    $payment_method->save();
  }

  public function refundPayment(PaymentInterface $payment, Price $amount = NULL) {
    define('FORTEECOM_DEMO_ENABLED', ($this->getMode() == 'test') ? 1 : 0);
    $this->assertPaymentState($payment, ['completed', 'partially_refunded']);
    // If not specified, refund the entire amount.
    $amount = $amount ?: $payment->getAmount();
    $this->assertRefundAmount($payment, $amount);

    $order                   = $payment->getOrder();
    $_forteecom_order_data   = $order->getData('commerce_forteecom');
    $_forteecom_order_id     = $_forteecom_order_data['_forteecom_order_id'];
    $_forteecom_session_id   = $_forteecom_order_data['_forteecom_session_id'];
    $_forteecom_merchant_id  = $_forteecom_order_data['_forteecom_merchant_id'];

    $refund_result = ForteEcomHelper::refund_payment(
      $_forteecom_merchant_id, 
      $_forteecom_order_id, 
      $_forteecom_session_id, 
      $amount->getNumber(), 
      $amount->getCurrencyCode()
    );
    if (!empty($refund_result['error_message'])) {
      $this->throwGatewayException('Could not refund payment');
    }

    $old_refunded_amount = $payment->getRefundedAmount();
    $new_refunded_amount = $old_refunded_amount->add($amount);
    if ($amount->lessThan($payment->getAmount())) {
      $payment->setState('partially_refunded');
    } else {
      $payment->setState('refunded');
    }

    $payment->setRefundedAmount($new_refunded_amount);
    $payment->save();
  }

  /**
   * @inheritDoc
   * @throws Drupal\Core\Entity\EntityStorageException
   */
  public function onReturn(OrderInterface $order, Request $request)
  {
    define('FORTEECOM_DEMO_ENABLED', ($this->getMode() == 'test') ? 1 : 0);

    $currency_code = $order->getTotalPrice()->getCurrencyCode();
    ForteEcomHelper::check_currency($currency_code);
    $messenger = Drupal::messenger();

    $_forteecom_order_data = $order->getData('commerce_forteecom');
    if (empty($_forteecom_order_data)) {
      throw new PaymentGatewayException('Missing ForteEcom order data');
    }

    $_forteecom_order_id    = $_forteecom_order_data['_forteecom_order_id'];
    $_forteecom_session_id  = $_forteecom_order_data['_forteecom_session_id'];
    $_forteecom_merchant_id = $_forteecom_order_data['_forteecom_merchant_id'];

    /** @var $payment PaymentInterface */
    $payment = null;

    $query       = Drupal::entityQuery('commerce_payment')->condition('remote_id', $_forteecom_order_id);
    $payment_ids = $query->execute();
    foreach ($payment_ids as $payment_id) {
      $payment = Payment::load($payment_id);

      $total             = $payment->getAmount()->getNumber();
      $order_id          = $payment->getOrderId();
      $order_description = Drupal::config('system.site')->get('name') . ' - Order: ' . $order_id;

      $payment_method_id = $request->request->getAlnum('forteecom_payment_method');
      if ($payment_method_id == 'new') {
        $status_result = ForteEcomHelper::get_payment_status($_forteecom_merchant_id, $_forteecom_order_id, $_forteecom_session_id);
      } else {
        /** @var PaymentMethodStorageInterface $payment_method_storage */
        $payment_method_storage = $this->entityTypeManager->getStorage('commerce_payment_method');

        /** @var PaymentMethod $payment_method */
        $payment_method = $payment_method_storage->load($payment_method_id);
        if (is_null($payment_method->getOwnerId()) || $payment_method->getOwnerId() != Drupal::currentUser()->id()) {
          // checking that payment method belongs to current user
          $this->throwGatewayException('Mismatched card data');
        }

        $status_result = ForteEcomHelper::purchase_by_token(
          $_forteecom_merchant_id,
          $_forteecom_order_id,
          $_forteecom_session_id,
          $total,
          $order_id,
          $order_description,
          $currency_code,
          $payment_method
        );

        if (!empty($status_result['error_message'])) {
          $payment->setState('failed');
          ForteEcomHelper::unset_order_forteecom_data($order);
          $payment->save();
          $this->throwGatewayException($status_result['error_message']);
        }
      }

      if (empty($status_result['error_message'])) {
        ForteEcomHelper::update_payment_status($order, $payment, $status_result['_forteecom_order_status']);
        ForteEcomHelper::process_payment_status($payment);
      } else {
        $messenger->addError('Error occurred while getting payment status');
      }
    }


  }

  /**
   * @param $message
   *
   * @throws PaymentGatewayException
   */
  private function throwGatewayException($message)
  {
    Drupal::messenger()->addError($message);
    throw new PaymentGatewayException($message);
  }

  /**
   * @inheritDoc
   */
  public function buildPaymentInstructions(PaymentInterface $payment)
  {
    $instruction_text = '';
    switch ($payment->getState()->getId()) {
      case 'holded':
        $instruction_text = 'Your payment is on-hold. Come back later to check its status.';
        break;
      case 'completed':
        $instruction_text = 'Your payment is completed.';
        break;
    }
    return [
      '#type' => 'processed_text',
      '#text' => $instruction_text,
    ];
  }

  /**
   * @inheritdoc
   * @throws Drupal\Core\Entity\EntityStorageException
   */
  public function deletePaymentMethod(PaymentMethodInterface $payment_method)
  {
    $payment_method->delete();
  }

  public function collectsBillingInformation()
  {
    return false;
  }

  /**
   * @inheritDoc
   */
  public function capturePayment(PaymentInterface $payment, Price $amount = NULL)
  {
    define('FORTEECOM_DEMO_ENABLED', ($this->getMode() == 'test') ? 1 : 0);
    $order                 = $payment->getOrder();
    $_forteecom_order_data = $order->getData('commerce_forteecom');

    $_forteecom_order_id     = $_forteecom_order_data['_forteecom_order_id'];
    $_forteecom_order_status = $_forteecom_order_data['_forteecom_order_status'];
    $_forteecom_session_id   = $_forteecom_order_data['_forteecom_session_id'];
    $_forteecom_merchant_id  = $_forteecom_order_data['_forteecom_merchant_id'];

    if ($_forteecom_order_status == ForteEcomHelper::STATUS_ON_HOLD) {
      //this should ever happen
      $this->throwGatewayException('Can not capture payment with status different from ON-HOLD');
    }

    $complete_result = ForteEcomHelper::complete_payment($_forteecom_merchant_id, $_forteecom_order_id, $_forteecom_session_id);
    if (!empty($complete_result['error_message'])) {
      $this->throwGatewayException('Error occurred while changing payment status from on-hold to complete');
    }

    $payment->setState('completed');
    $payment->setRemoteState(ForteEcomHelper::STATUS_SUCCESS);
    $payment->setAmount($payment->getAmount());

    $payment->save();
  }

  /**
   * @inheritDoc
   */
  public function voidPayment(PaymentInterface $payment)
  {
    $this->throwGatewayException('Currently ForteEcom does not support VOID operations');
  }

  /**
   * Creates a payment.
   *
   * @param PaymentInterface $payment
   *   The payment.
   * @param bool             $capture
   *   Whether the created payment should be captured (VS authorized only).
   *   Allowed to be FALSE only if the plugin supports authorizations.
   *
   * @throws InvalidArgumentException
   *   If $capture is FALSE but the plugin does not support authorizations.
   * @throws PaymentGatewayException
   *   Thrown when the transaction fails for any reason.
   */
  public function createPayment(PaymentInterface $payment, $capture = TRUE)
  {
    $currency_code = $payment->getAmount()->getCurrencyCode();
    ForteEcomHelper::check_currency($currency_code);

    $order             = $payment->getOrder();
    $messenger         = Drupal::messenger();
    $total             = $payment->getAmount()->getNumber();
    $order_id          = $payment->getOrderId();
    $order_description = Drupal::config('system.site')->get('name') . ' - Order: ' . $order_id;

    $configuration = $this->getConfiguration();

    $init_result = ForteEcomHelper::init_payment(
      $configuration['merchant_id'],
      $configuration['pre_auth'],
      $total,
      $order_id,
      $order_description,
      $currency_code
    );

    if (!empty($init_result['error_message'])) {
      $this->throwGatewayException('Could not initialize ForteEcom payment');
    }

    $_forteecom_order_id    = $init_result['_forteecom_order_id'];
    $_forteecom_session_id  = $init_result['_forteecom_session_id'];
    $_forteecom_merchant_id = $configuration['merchant_id'];

    $order->setData('commerce_forteecom', [
      '_forteecom_order_id'     => $_forteecom_order_id,
      '_forteecom_order_status' => ForteEcomHelper::STATUS_INIT,
      '_forteecom_session_id'   => $_forteecom_session_id,
      '_forteecom_merchant_id'  => $_forteecom_merchant_id,
    ]);

    $purchase_result = ForteEcomHelper::purchase_by_token(
      $_forteecom_merchant_id,
      $_forteecom_order_id,
      $_forteecom_session_id,
      $total,
      $order_id,
      $order_description,
      $currency_code,
      $payment->getPaymentMethod()
    );

    if (!empty($purchase_result['error_message'])) {
      $this->throwGatewayException($purchase_result['error_message']);
    }

    ForteEcomHelper::process_payment_status($payment);

    $order->save();
    $payment->setRemoteId($_forteecom_order_id);
    $payment->setState('inited');
    $payment->setRemoteState(ForteEcomHelper::STATUS_INIT);
    $payment->save();

  }
}
